-- Attempts at alleviating equality reasoning
-- Author :: Ulf Norell
-- Created:: 2004-12-15

-- # import Property
-- # import Equality



-- Steps ------------------------------------------------------------------

--EqStep : (A:Set) -> (x:A) -> Type
--EqStep A x = { y : A; Eq A x y }

data EqStep (A:Set)(x:A) : Prop where
  by    : (y : A) -> x==y -> EqStep A x

infix 14 `by`

eqStepRHS : (A:Set) -> (x:A) -> EqStep A x -> A
eqStepRHS A x (y `by` p) = y

eqStepProof : (A:Set) -> (x:A) -> (s : EqStep A x) ->
                x == eqStepRHS A x s
eqStepProof A x (y `by` p) = p

-- Attempt 1 --------------------------------------------------------------
{-
eqStep :  (A:Set) -> (x : A)                  ->
          (s1 : EqStep A x)                   ->
          (s2 : EqStep A (eqStepRHS A x s1)) ->
          EqStep A x
eqStep A x s1 s2 = z `by` transitivity A x y z xy yz
  where y   = eqStepRHS A x s1
        z   = eqStepRHS A y s2
        xy  = eqStepProof A x s1
        yz  = eqStepProof A y s2

refStep : (A:Set) -> (x:A) -> EqStep A x
refStep A x = x `by` reflexivity A x
-}
-- Attempt 2 --------------------------------------------------------------

EqProof : (A:Set) -> Type
EqProof A = sig { lhs, rhs : A; proof : lhs==rhs }

beginEqProof : (A:Set) -> (x:A) -> EqProof A
beginEqProof A x = struct {lhs = x; rhs = x; proof = reflexivity x}

extractProof : (A:Set) -> (p : EqProof A) -> p.lhs == p.rhs
extractProof A p = p.proof

eqStep : (A:Set) -> (p : EqProof A) -> EqStep A (p.rhs) -> EqProof A
eqStep A p s = struct 
                 lhs   = x
                 rhs   = z
                 proof = impE (impE (transitivity x y z) p.proof) 
                              (eqStepProof A y s)
  where
    x = p.lhs
    y = p.rhs
    z = eqStepRHS A y s

-- Attempt 3 --------------------------------------------------------------
-- Completely abstract, doesn't work

{-
data EqStep (A:Set) : Set where
  by  : (x:A) -> Eq A ? x -> EqStep A

data EqChain (A:Set) : Set where
  init  : (x:A) -> EqChain A
  (===) : EqChain A -> EqStep A -> EqChain A
-}
-- Requires induction recursion
{-
data EqProof' (A:Set) : Prop where
  (===) : EqProof' A -> EqStep A lhs -> EqProof' A
-}

