
module Compilers.Haskell where

import Control.Monad.State
import Data.Char
import Language.Haskell.Syntax
import Language.Haskell.Pretty

import Interface.Options

import Syntax.Internal
import Syntax.Pretty
import Syntax.Position

import TypeChecker.Monad
import TypeChecker.Monad.Options
import TypeChecker.Monad.Definitions
import TypeChecker.Values
import TypeChecker.UnfoldConstants
import TypeChecker.Abstractions
import TypeChecker.Patterns

import Utils.Map
import Utils.Monads
import Utils.Debug

compile :: TC HsModule
compile = undefined

{-
tick :: State Int Int
tick = do   n <- get
	    put (n + 1)
	    return n

freshName :: String -> State Int String
freshName s = tick >>= \n -> return $ s ++ show n

freshVar = HsIdent <$> freshName "x"

compile :: TC HsModule
compile =
    do	file	    <- last <$> (optInputFiles <$> commandLineOptions)
	defs	    <- getDefinitions
	theDatatype <- createTheDatatype
	let hsDefs  = concatMap (uncurry compileDef) $ assocsMap defs
	return $ HsModule   noLoc
			    (Module $ moduleName file)
			    Nothing
			    [HsImportDecl   noLoc
					    (Module "Compiler.Prelude")
					    False
					    Nothing
					    Nothing
			    ]
			    (theDatatype ++ hsDefs)

moduleName = hUpper . takeWhile (/='.') . reverse . takeWhile (/='/') . reverse
    where
	hUpper (c:s) = toUpper c : s

noLoc = SrcLoc "" 0 0

getAxioms :: TC [(Name, PolymorphicDefValue)]
getAxioms =
    do	defs <- assocsMap <$> getDefinitions
	return [ def | def@(x, PolyDefVal _ _ _ [] _) <- defs ]

createTheDatatype :: TC [HsDecl]
createTheDatatype =
    do	axs <- getAxioms
	let cons = map axiomToConDecl axs
	return
	    [ HsTypeDecl noLoc hsAgda []
			    ( HsTyCon hsQAgdaValue `HsTyApp`
			      HsTyCon hsQAgdaConst
			    )
	    , HsDataDecl noLoc [] hsAgdaConst [] cons
			{- deriving -} [UnQual $ HsIdent "Show"]
	    ]

getArity :: Type -> Int
getArity t =
    case unfoldConstants t of
	VPi w f	_
	    | isSort w	-> getArity (f `clApp` VId "_" w)
	    | otherwise	-> 1 + getArity (f `clApp` VId "_" w)
	_		-> 0

dropLambdas :: Type -> Type
dropLambdas t =
    case unfoldConstants t of
	VLam f	-> dropLambdas $ f `clApp` VId "_" (clType f)
	_   -> t

polyDefArity (PolyDefVal _ _ _ _ t)   = getArity $ dropLambdas t
defArity (DefVal _ _ _ _ t)	    = getArity $ dropLambdas t

axiomToConDecl :: (Name, PolymorphicDefValue) -> HsConDecl
axiomToConDecl (x, d) =
	HsConDecl noLoc (hsCon x) args
    where
	args = replicate (polyDefArity d) hsBangedAgdaTy

compileDef :: Name -> PolymorphicDefValue -> [HsDecl]
compileDef x d@(PolyDefVal _ _ _ cs t)
    | isSort t	= []
    | otherwise	=
	case cs of
	    []	->
		[ HsFunBind  [ HsMatch  noLoc (hsName x) ps
					(HsUnGuardedRhs $
					    HsCon constQCon `HsApp`
					    (HsCon (hsQCon x) `hsApps` vs)
					)
					[]
			    ]
		]
		where
		    xs = [ HsIdent $ "x" ++ show n | n <- [1..ar] ]
		    ps = map HsPVar xs
		    vs = map (HsVar . UnQual) xs
	    cs	-> [ HsFunBind $ map (compileClause x t) cs ]
	where
	    ar = polyDefArity d

compileClause :: Name -> Type -> ClauseValue -> HsMatch
compileClause x t cl@(ClauseVal ps v) =
    HsMatch noLoc (hsName x)
	    (concatMap compilePattern vs) rhs []
    where
	ps'	= ps ++ drop (length ps) (vars t)
	vs	= undefined --TODO: map evaluateTypedPattern ps'
	xs	= zipWith VId (concatMap patternVars ps') (repeat undefined)
	rhs'	= head $ compileValue $ v `vApps` xs
	rhs	= HsUnGuardedRhs rhs'
	vars t	= case unfoldConstants t of
		    VPi w f _	-> VarP noRange x : vars (f `clApp` VId x w)
			where x = clName f
		    _		-> []

compilePattern :: Value -> [HsPat]
compilePattern p
    | isType p	= []
    | otherwise	=
	case vAppView p of
	    VAVId "_" _ []  -> [HsPWildCard]
	    VAVId x _ []    -> [HsPVar (hsName x)]
	    VAVDef c _ ps   ->
		[ HsPParen $ HsPApp constQCon
		    [ HsPParen	$ HsPApp (hsQCon c)
				$ concatMap compilePattern ps
		    ]
		]
	    VAVPair p q _   ->
		case concatMap compilePattern [p,q] of
		    ps@[_,_]	-> [ HsPParen $ HsPApp pairQCon ps ]
		    ps		-> ps

compileValue :: Value -> [HsExp]
compileValue v = evalState (compileV v) 0
    where
	compileV v
	    | isType v	= return []
	    | otherwise	=
		case vAppView v of

		    VAVDef x d vs   ->
			do  vs' <- concatMapM compileV vs
			    let n = length vs'
			    if (n < ar)
				then 
				    -- Make sure constants are always fully applied
				    do  xs <- replicateM (ar - n) freshVar
					let ps  = map HsPVar xs
					    es  = map (HsVar . UnQual) xs
					return [ vLambda ps $ f `hsApps` (vs' ++ es) ]
				else
				    do  let (vs0,vs1) = splitAt ar vs'
					return [ f `hsApps` vs0 `hsVApps` vs1 ]
			where
			    ar  = defArity d
			    f   = HsVar (hsQName x)

		    VAVId x _ vs	->
			do  es <- concatMapM compileV vs
			    return [ HsVar (hsQName x) `hsVApps` es ]

		    VAVLam f	->
			do  x <- freshName (clName f)
			    e <- compileV (f `clApp` VId x (clType f))
			    return $ map (vLambda [HsPVar $ hsName x]) e
			    

		    VAVProj v p vs	    ->
			do  e <- compileV v
			    es <- concatMapM compileV vs
			    let pr = case p of
					VFst	-> fstVar
					VSnd	-> sndVar
			    return $ map (\e -> pr `hsApps` [e] `hsVApps` es) e

		    VAVPair v w _	    ->
			do  es <- concatMapM compileV [v,w]
			    return $ case es of
				[_,_]   -> [ HsCon pairQCon `hsApps` es ]
				_	    -> es

		    VAVPi _ _ _		-> return [ neverMind ("{" ++ show v ++ "}") ]
		    VAVSigma _ _ _ _ _	-> return [ neverMind ("{" ++ show v ++ "}") ]
		    VAVTop _		-> return [ neverMind ("{" ++ show v ++ "}") ]

isType :: Value -> Bool
isType v = isSort $ valueType v

isSort :: Type -> Bool
isSort v = case unfoldConstants v of
		VSort _	    -> True
		VPi w f _   -> isSort (f `clApp` VId (clName f) w)
		_	    -> False

vLambda :: [HsPat] -> HsExp -> HsExp
vLambda = flip (foldr lam)
    where
	lam p@(HsPVar x) e =
	    HsCon funQCon `hsApps`  [ HsLit $ HsString $ nameToString x
				    , HsParen $ HsLambda noLoc [p] e
				    ]

hsVApps v [] = v
hsVApps v vs = HsParen $ foldl vapp v vs
    where
	vapp v w = HsVar appQName `hsApps` [v, w]

hsApps v [] = v
hsApps v vs = HsParen $ foldl HsApp v vs

pairQCon	    = UnQual pairCon
pairCon		    = HsIdent "Pair"
funQCon		    = UnQual funCon
funCon		    = HsIdent "Lam"
neverMind x	    = HsCon neverMindQCon `hsApps` [ HsLit $ HsString x ]
neverMindQCon	    = UnQual neverMindCon
neverMindCon	    = HsIdent "NeverMind"
appQCon		    = UnQual appCon
appCon		    = HsIdent "App"
fstQCon		    = UnQual fstCon
fstCon		    = HsIdent "Fst"
sndQCon		    = UnQual sndCon
sndCon		    = HsIdent "Snd"
constQCon	    = UnQual constCon
constCon	    = HsIdent "NewConstant"

appQName	    = UnQual appName
appName		    = HsIdent "app"
fstVar		    = HsVar fstQName
fstQName	    = UnQual fstName
fstName		    = HsIdent "first"
sndVar		    = HsVar sndQName
sndQName	    = UnQual sndName
sndName		    = HsIdent "second"

hsQCon	= UnQual . hsCon
hsQName = UnQual . hsName

hsBangedAgdaTy  = HsUnBangedTy hsAgdaTy
hsAgdaTy	= HsTyCon hsQAgda

hsQAgdaConst	= UnQual hsAgdaConst
hsAgdaConst	= HsIdent "AgdaConstant"

hsQAgdaValue	= UnQual hsAgdaValue
hsAgdaValue	= HsIdent "AgdaValue"

hsQAgda		= UnQual hsAgda
hsAgda		= HsIdent "Agda"

hsCon x
    | isOperator x  = HsSymbol $ ":~" ++ x
    | otherwise	    = HsIdent $ "C" ++ x

hsName x
    | isOperator x  = HsSymbol $ "~" ++ x
    | otherwise	    = HsIdent $ "v" ++ x

nameToString (HsIdent x) = x
nameToString (HsSymbol x) = x
-}
