{-# OPTIONS -cpp #-}

{-| Import chasing. Requires the variable @_ROOT_@ to be cpp'd to the root
    directoty of the agdaLight distribution.
-}
module Interface.ChaseImports where

import Control.Monad
import Data.Char
import Data.List
import System.Directory

import Utils.Monads
import Utils.Files

-- | Find a file.
findModule :: [FilePath] -> String -> [String] -> IO (Maybe FilePath)
findModule inc name exts =
    do	cwd <- addSlash <$> getCurrentDirectory
	let paths = "" : map addSlash inc
	mf <- findFileName paths name exts
	return $ fmap (canonify cwd) mf

{-|
    @chaseImports include_dirs files@.
    Chase imports by looking for lines of the form @-- # import foo.bar.baz@
    The interpretation is that there should be a file foo\/bar\/baz.agda in
    either the same directory as the file containing the import, in the current
    working directory, or in the specified include directories (agdaLight\/lib
    is default).
-}
chaseImports :: [FilePath] -> [FilePath] -> IO [FilePath]
chaseImports is files =
    do	cwd <- addSlash <$> getCurrentDirectory
	(nub . concat)
	    <$> mapM (\file -> findImports cwd includes [file]) files
    where
	includes = map addSlash is

findImports :: FilePath -> [FilePath] -> [FilePath] -> IO [FilePath]
findImports _ _ [] = error "impossible Interface.ChaseImports.findImports []"
findImports cwd includes prev_files@(file:_)
    | isCycle prev_files    = fail $ "While chasing imports: Cycle detected:\n"
				    ++ unlines (map ("  " ++ ) prev_files)
    | otherwise		    =
	do  imps	<- (concatMap findImport . lines) <$> readFile file
	    let paths = getPath file : "" : includes
	    files	<- map (canonify cwd) <$> mapM (getFileName paths) imps
	    all_files	<- concat <$>
			    mapM (\file -> findImports cwd includes (file:prev_files))
				files
	    return $ all_files ++ [file]
    where
	findImport s =
	    do	("--", s0)	<- lex s
		("#", s1)	<- lex s0
		("import", s2)	<- lex s1
		return $ takeWhile (not . isSpace) $ dropWhile isSpace s2

isCycle xs = nub xs /= xs

getPath :: FilePath -> FilePath
getPath file = reverse $ dropWhile (/=slash) $ reverse file

getFileName paths imp =
    do	mf <- findFileName paths imp [".agda",".lagda"]
	case mf of
	    Nothing -> fail $ "While chasing imports: File not found " ++ imp
	    Just x  -> return x

findFileName :: [FilePath] -> String -> [String] -> IO (Maybe FilePath)
findFileName paths imp exts =
    foldr (\file getFN ->
		ifM (doesFileExist file)
		    (return $ Just file)
		    getFN
	)   (return Nothing)
	    (concatMap (\path -> map (path++) files) paths)
    where
	files	= [ map unDot imp ++ ext | ext <- exts ]

	unDot '.'   = slash
	unDot c	    = c

