{-| Pretty print a FOL formula in a form suitable for santa.
    TODO: Doesn't work at the moment.
-}
module Plugins.FOL.Santa where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.State
import Data.Char
import Data.List
import Text.PrettyPrint

import Syntax.Name

import Utils.Map hiding (map)
import Utils.Names
import Utils.Monads

import Plugins.FOL.Types
import Plugins.FOL.Primitive

data SantaType = Axiom | Hypothesis | Conjecture | Negated_Conjecture

instance Show SantaType where
    show Axiom		    = "axiom"
    show Hypothesis	    = "hypothesis"
    show Conjecture	    = "conjecture"
    show Negated_Conjecture = "negated_Conjecture"

-- | TODO: figure out what the tag is.
makeSantaFile :: [String] -> [Formula] -> Formula -> String
makeSantaFile includes pre goal =
    unlines (  map santaInclude includes
	    ++ [folToSanta "goal" Negated_Conjecture (Not goal)]
	    ++ map (folToSanta "premiss" Axiom) pre
	    )
    where
	santaInclude incl = "include ('" ++ incl ++ "')."

type M = State UsedNames

runM m = evalState m noNamesUsed

freshName :: String -> M String
freshName x = santaVar <$> fresh x

-- | TODO: do this nicely!
folToSanta :: String -> SantaType -> Formula -> String
folToSanta tag t f 
  = "fof(" ++ tag ++ "," ++ map toLower (show t) ++ "," ++ runM (body f) ++ ").\n"
  where
    body :: Formula -> M String
    body z 
          = case z of 
	      Predicate "equal" [x,y] ->
		do  let [s,t] = map termToSanta [x,y]
		    return $ "equal(" ++ s ++ "," ++ t ++ ")"
              Predicate v xs ->
		do  let ts = map termToSanta xs
		    return $ santaConst v ++ "(" ++
                                  concatMap id (intersperse "," ts) ++ ")"
              And g h        -> binOpSanta "&" g h
              Or  g h        -> binOpSanta "|" g h
              Not h          ->
		do  t <- body h
		    return $ "~ (" ++ t ++ ")"
              Implies g h    -> binOpSanta "=>" g h
              Equiv g h      -> binOpSanta "<=>" g h
              ForAll x f     -> quantSanta "!" x f
              Exists x f     -> quantSanta "?" x f
              TRUE           -> return "cANY => cANY"	    --"$true"
              FALSE          -> return "~(cANY => cANY)"    --"$false"

    binOpSanta :: String -> Formula -> Formula -> M String
    binOpSanta op g h =
	do  t1 <- body g
	    t2 <- body h
	    return $ "(" ++ t1 ++ ")" ++ op ++ "(" ++ t2 ++ ")"

    quantSanta :: String -> String -> (FOLTerm -> Formula) -> M String
    quantSanta q v f =
	do  x <- freshName v
	    t <- body (f $ FOLVar x)
	    return $ q ++ " ["  ++ x ++ "]: (" ++ t ++ ")"

requote :: String -> String -> String
requote q ('"':x)   = q ++ init x ++ q
requote q x	    = q ++ x ++ q

santaVar (Name _ x n)	= requote "\"" $ quoted (capitalize . tag n) x
santaVar DontCare	= error "impossible: Plugins.FOL.Santa.santaVar DontCare"

santaConst x = requote "\"" $ quoted uncapitalize x

tag n s = s ++ "_" ++ show n

capitalize x	= 'V':x
uncapitalize "_|_"  = "_|_"
uncapitalize x	    = 'c':x

termToSanta :: FOLTerm -> String
termToSanta t 
  = case t of
      FOLFun n [t1,t2]
	| n == kwApp -> termToSanta t1 ++ "[" ++ termToSanta t2 ++ "]"
      FOLFun n []   -> santaConst n
      FOLFun n xs   ->  let ts = concatMap id $ intersperse "," $ map termToSanta xs
			in santaConst n ++ "(" ++ ts ++ ")"
      FOLVar n    -> n

